/*
 * Renderer 2. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import javax.swing.Timer;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.io.File;

/**
   This program creates an animation by repeatedly
   drawing to the Java graphics window.
*/
public class Animation
{
   private static final String assets = "../assets/";

   public static double[] deltaX, deltaY, deltaZ;

   public Animation()
   {
      // Create the Scene object that we shall render
      final Scene scene = new Scene("Animation");

      // Create several Model objects.
      final Model m1 = new ObjSimpleModel(new File(assets + "apple.obj"));
      final Model m2 = new ObjSimpleModel(new File(assets + "cow.obj"));
      final Model m3 = new ObjSimpleModel(new File(assets + "cessna.obj"));
      final Model m4 = new ObjSimpleModel(new File(assets + "galleon.obj"));
      final Model m5 = new ObjSimpleModel(new File(assets + "teapot.obj"));
      final Model m6 = new ObjSimpleModel(new File(
                          assets + "small_rhombicosidodecahedron.obj"));
      final Model m7 = new ObjSimpleModel(new File(
                          assets + "great_rhombicosidodecahedron.obj"));

      // Add some Models to the Scene.
      scene.addPosition(new Position(m1),  // apple
                        new Position(m3),  // cessna
                        new Position(m4),  // galleon
                        new Position(m2)); // cow

      deltaX = new double[scene.positionList.size()];
      deltaY = new double[scene.positionList.size()];
      deltaZ = new double[scene.positionList.size()];

      // Push the models away from where the camera is.
      for (int i = 0; i < scene.positionList.size(); ++i)
      {
         deltaX[i] =  0.0;
         deltaY[i] =  0.0;
         deltaZ[i] = -3.0;
         scene.getPosition(i).translation(0.0, 0.0, -3.0);
      }

      Rasterize.doClipping = false; // turn off clipping

      // Create a FrameBufferPanel that holds a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;
      final FrameBufferPanel fbp = new FrameBufferPanel(width, height,
                                                        Color.darkGray);

      // Create a JFrame that will hold the FrameBufferPanel.
      final JFrame jf = new JFrame("Renderer 2");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setLocationRelativeTo(null);
      jf.setResizable(false);
      jf.setVisible(true);

      // Create an ActionListener event handler object that will respond
      // to Timer events and redraw the Scene into the FrameBufferPanel.
      ActionListener actionListener = new ActionListener()
      {
         // Implement the ActionListener interface (one method).
         @Override public void actionPerformed(ActionEvent e)
         {
            // Update the Scene.
            deltaX[0] +=  0.3;  // move right
            deltaY[0] +=  0.1;  // move up
            deltaZ[0] += -0.1;  // move back
            scene.getPosition(0).translation(deltaX[0],
                                             deltaY[0],
                                             deltaZ[0]);
            deltaX[1] += -0.3;  // move left
            deltaY[1] += -0.1;  // move down
            deltaZ[1] += -0.1;  // move back
            scene.getPosition(1).translation(deltaX[1],
                                             deltaY[1],
                                             deltaZ[1]);
            deltaX[2] +=  0.01;  // move right
            scene.getPosition(2).translation(deltaX[2],
                                             deltaY[2],
                                             deltaZ[2]);
            // Render again.
            final FrameBuffer fb = fbp.getFrameBuffer();
            fb.clearFB();
            Pipeline.render(scene, fb);
            fbp.update();
         }
      };

      // Create a Timer object that will generate ActionEvent
      // objects for the ActionListener object to respond to.
      final int fps = 30;
      final Timer timer = new Timer(1000/fps, actionListener);
      timer.start();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new Animation()
      );
   }
}
