/*
 * Renderer 2. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.util.PointCloud;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;
import java.io.File;

/**

*/
public class InteractiveGRSModels_R2 implements KeyListener,
                                                ComponentListener
{
   private static final String assets = "../assets/";

   private final Scene scene;  // The event handlers need
   private final JFrame jf;    // access to these fields.
   private final FrameBufferPanel fbp;
   private int currentModel = 0;
   private double deltaX = 0.0;
   private double deltaY = 0.0;
   private double deltaZ = 0.0;
   private boolean displayTranslation = false;
   private Model savedModel = null; // used to hold a PointCloud model
   private int pointSize = 0;       // used by the point clouds

   private boolean takeScreenshot = false;
   private int screenshotNumber = 0;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public InteractiveGRSModels_R2()
   {
      // Create the Scene object that we shall render.
      scene = new Scene();

      // Instantiate all the grs models.
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/bronto_v2.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/rex_v2.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/usa_v2.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/bronto.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/rex.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/usa.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/vinci.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/dragon.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/birdhead.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/knight.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/house.grs"))));
      scene.addPosition(new Position(new GRSModel(
                             new File(assets + "grs/scene.grs"))));

      // Create a set of x and y axes.
      final Model axes = new Axes2D(-1, +1, -1, +1, 20, 20);
      final Position axes_p = new Position(axes);
      // Have the axes pushed away from where the camera is.
      axes_p.translation(0, 0, -2);
      scene.addPosition(axes_p);

      // Make the interactive models invisible, except for the current model.
      for (int i = 0; i <  scene.positionList.size() - 1; ++i)
      {
         scene.getPosition(i).visible = false;
      }
      currentModel = 0; // bronto_v2
      scene.getPosition(currentModel).visible = true;

      // Have the models pushed away from where the camera is.
      deltaX =  0;
      deltaY =  0;
      deltaZ = -2;
      scene.getPosition(currentModel).translation(deltaX, deltaY, deltaZ);

      // Switch to a parallel (orthographic) projection.
      scene.getCamera().perspective = false;

      Rasterize.doClipping = true;


      // Define initial dimensions for a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;

      // Create a FrameBufferPanel that holds a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 2");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setLocationRelativeTo(null);
      jf.setVisible(true);

      // Register this object as the event listener for JFrame events.
      jf.addKeyListener(this);
      jf.addComponentListener(this);

      print_help_message();
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e){}
   @Override public void keyReleased(KeyEvent e){}
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c && e.isAltDown())
      {
         System.out.println();
         System.out.println(scene.getPosition(currentModel).getModel());
      }
      else if ('d' == c)
      {
         // debug just the current model
         final Position p = scene.getPosition(currentModel);
         p.debug = ! p.debug;
      }
      else if ('D' == c)
      {
         Rasterize.debug = ! Rasterize.debug;
      }
      else if ('c' == c)
      {
         Rasterize.doClipping = ! Rasterize.doClipping;
         System.out.print("Clipping is turned ");
         System.out.println(Rasterize.doClipping ? "On" : "Off");
      }
      else if ('/' == c)
      {
         scene.getPosition(currentModel).visible = false;
         currentModel = (currentModel + 1) % (scene.positionList.size() - 1);
         scene.getPosition(currentModel).visible = true;
         savedModel = null;
         pointSize = 0;
      }
      else if ('?' == c)
      {
         scene.getPosition(currentModel).visible = false;
         currentModel = currentModel - 1;
         if (currentModel < 0) currentModel = scene.positionList.size() - 2;
         scene.getPosition(currentModel).visible = true;
         savedModel = null;
         pointSize = 0;
      }
      else if ('p' == c)
      {
         scene.getCamera().perspective = ! scene.getCamera().perspective;
         final String p = scene.getCamera().perspective ? "perspective"
                                                        : "orthographic";
         System.out.println("Using " + p + " projection");
      }
      else if ('P' == c)
      {
         if (savedModel != null)
         {
            scene.getPosition(currentModel).setModel(savedModel);
            savedModel = null;
            ++pointSize;
         }
         else
         {
            final Model model = scene.getPosition(currentModel).getModel();
            savedModel = model;
            scene.getPosition(currentModel).setModel(
                        PointCloud.make(model, pointSize));
         }
      }
      else if ('m' == c) //display transformation information
      {
         displayTranslation = ! displayTranslation;
      }
      else if ('=' == c) // Reset the translation vector.
      {
         deltaX =  0;
         deltaY =  0;
         deltaZ = -2;
      }
      else if ('x' == c) // Translate the cube left.
      {
         deltaX += -0.1;
      }
      else if ('X' == c) // right
      {
         deltaX += +0.1;
      }
      else if ('y' == c) // down
      {
         deltaY += -0.1;
      }
      else if ('Y' == c)  // up
      {
         deltaY += +0.1;
      }
      else if ('z' == c)  // back
      {
         deltaZ += -0.1;
      }
      else if ('Z' == c)// forward
      {
         deltaZ += +0.1;
      }
      else if ('+' == c)
      {
         takeScreenshot = true;
      }

      scene.getPosition(currentModel).translation(deltaX, deltaY, deltaZ);

      if (displayTranslation && ('m'==c||'='==c
                               ||'x'==c||'y'==c||'z'==c
                               ||'X'==c||'Y'==c||'Z'==c))
      {
         System.out.printf("deltaX = % .8f, deltaY = % .8f, " +
                           "deltaZ = % .8f\n",
                            deltaX, deltaY, deltaZ);
      }

      // Render again.
      final FrameBuffer fb = fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb);
      if (takeScreenshot)
      {
         fb.dumpFB2File(String.format("Screenshot%03d.png", screenshotNumber),
                        "png");
         ++screenshotNumber;
         takeScreenshot = false;
      }
      fbp.update();
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        jf.getWidth(), jf.getHeight(),
                        fbp.getWidth(), fbp.getHeight());

      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final FrameBuffer fb = new FrameBuffer(w, h);
      fbp.setFrameBuffer(fb);
      Pipeline.render(scene, fb);
      fbp.update();
   }


   private void print_help_message()
   {
      System.out.println("Use the 'd/D' keys to toggle debugging information on and off for the current model.");
      System.out.println("Use the '/' and '?' keys to cycle forwards and backwards through the models.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the models along the x, y, z axes.");
      System.out.println("Use the 'm' key to toggle the display of translation information.");
      System.out.println("Use the '=' key to reset the model translation.");
      System.out.println("Use the 'c' key to toggle line clipping on and off.");
      System.out.println("Use the 'P' key to convert the current model to a point cloud.");
      System.out.println("Use the '+' key to save a \"screenshot\" of the framebuffer.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new InteractiveGRSModels_R2()
      );
   }
}
