/*
 * Renderer 2. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.util.PointCloud;
import renderer.scene.util.MeshMaker;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;
import java.io.File;

/**

*/
public class InteractiveModelsAll_R2 implements KeyListener,
                                                ComponentListener
{
   private static final String assets = "../assets/";

   private final Scene scene;  // The event handlers need
   private final JFrame jf;    // access to these fields.
   private final FrameBufferPanel fbp;
   private int currentModel = 0;
   private double deltaX = 0.0;
   private double deltaY = 0.0;
   private double deltaZ = 0.0;
   private boolean displayTranslation = false;
   private Model savedModel = null; // used to hold a PointCloud model
   private int pointSize = 0;       // used by the point clouds

   private boolean takeScreenshot = false;
   private int screenshotNumber = 0;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public InteractiveModelsAll_R2()
   {
      // Create the Scene object that we shall render
      scene = new Scene();

      // Instantiate at least one of every Model class.
      // 2D models
      scene.addPosition(new Position(
                           new Square(1.0)));
      scene.addPosition(new Position(
                           new SquareGrid(1.0, 11, 15)));
      scene.addPosition(new Position(
                           new Circle(1.0, 16)));
      scene.addPosition(new Position(
                           new CircleSector(1.0,  Math.PI/4, -Math.PI/4, 13)));
      scene.addPosition(new Position(
                           new CircleSector(1.0, -Math.PI/4,  Math.PI/4,  5)));
      scene.addPosition(new Position(
                           new Disk(1.0, 4, 16)));
      scene.addPosition(new Position(
                           new DiskSector(1.0, Math.PI/2, 3*Math.PI/2, 4, 8)));
      scene.addPosition(new Position(
                           new Ring(1.0, 0.25, 3, 16)));
      scene.addPosition(new Position(
                           new RingSector(1.0, 0.25, Math.PI/2, 3*Math.PI/2, 3, 8)));
      scene.addPosition(new Position(
                           new BarycentricTriangle(45, 4))); // try 5 or 6
      // cubes
      scene.addPosition(new Position(new Box(1.0, 1.0, 1.0)));
      scene.addPosition(new Position(new Cube()));
      scene.addPosition(new Position(new Cube2(4, 5, 6)));
      scene.addPosition(new Position(new Cube2(40, 50, 60)));
      scene.addPosition(new Position(new Cube3(12, 14, 15)));
      scene.addPosition(new Position(new Cube4(12, 14, 15)));
      // polyhedra
      scene.addPosition(new Position(new Tetrahedron()));
      scene.addPosition(new Position(new Tetrahedron(true)));
      scene.addPosition(new Position(new Tetrahedron(20, 40)));
      scene.addPosition(new Position(new Tetrahedron(30, 30, 30, 30)));
      scene.addPosition(new Position(new Octahedron()));
      scene.addPosition(new Position(new Octahedron(30)));
      scene.addPosition(new Position(new Octahedron(20, 20, 20)));
      scene.addPosition(new Position(new Icosahedron()));
      scene.addPosition(new Position(new Dodecahedron()));
      // pyramids
      scene.addPosition(new Position(new Pyramid(
                2.0, 1.0, 5, 6)));
      scene.addPosition(new Position(new PyramidFrustum(
                2.0, 1.0, 0.5, 4, 5)));
      scene.addPosition(new Position(new PyramidFrustum(
                1.0, 2.0, 0.5, 4, 5)));
      scene.addPosition(new Position(new TriangularPyramid(
                Math.sqrt(3)/Math.sqrt(2))));
      scene.addPosition(new Position(new TriangularPyramid(
                1.0, 1.0, 7, 7)));
      scene.addPosition(new Position(new TriangularPrism(
                0.6, 0.5, 0.5, 3, true)));
      scene.addPosition(new Position(new ViewFrustumModel()));
      // cones
      scene.addPosition(new Position(new Cone(
                1.0, 1.0, 10, 16)));
      scene.addPosition(new Position(new ConeSector(
                1.0, 1.0, 0.5, 0, 2*Math.PI, 5, 16)));
      scene.addPosition(new Position(new ConeSector(
                1.0, 1.0, 0.5, Math.PI/2, 3*Math.PI/2, 5, 8)));
      scene.addPosition(new Position(new ConeFrustum(
                1.0, 0.5, 0.5, 6, 16)));
      scene.addPosition(new Position(new ConeFrustum(
                0.5, 1.0, 0.5, 6, 16)));
      // cylinders
      scene.addPosition(new Position(new Cylinder(
                0.5, 1.0, 11, 12)));
      scene.addPosition(new Position(new CylinderSector(
                0.5, 1.0, Math.PI/2, 3*Math.PI/2, 11, 6)));
      // spheres
      scene.addPosition(new Position(new Sphere(
                1.0, 15, 12)));
      scene.addPosition(new Position(PointCloud.make(new Sphere(
                1.0, 60, 60))));
      scene.addPosition(new Position(PointCloud.make(new Sphere(
                1.0, 120, 120))));
      scene.addPosition(new Position(new SphereSector(
                1.0, Math.PI/2, 3*Math.PI/2, Math.PI/4, 3*Math.PI/4, 7, 6)));
      scene.addPosition(new Position(new SphereSubdivided(4)));
      scene.addPosition(new Position(new SphereSubdivided(6, true, false)));
      scene.addPosition(new Position(new SphereSubdivided(7, false, true)));
      // torus
      scene.addPosition(new Position(new Torus(
                0.75, 0.25, 12, 16)));
      scene.addPosition(new Position(PointCloud.make(new Torus(
                0.75, 0.25, 120, 120))));
      scene.addPosition(new Position(new TorusSector(
                0.75, 0.25, Math.PI/2, 3*Math.PI/2, 12, 8)));
      scene.addPosition(new Position(new TorusSector(
                0.75, 0.25, 0, 2*Math.PI, Math.PI, 2*Math.PI, 6, 16)));
      scene.addPosition(new Position(new TorusSector(
                0.75, 0.25, 0, 2*Math.PI, -Math.PI/2, Math.PI/2, 6, 16)));
      scene.addPosition(new Position(new TorusSector(
                0.75, 0.25, Math.PI/2, 3*Math.PI/2, -Math.PI/2, Math.PI/2, 6, 8)));
      // model files
      scene.addPosition(new Position(new GRSModel(
                new File(assets + "grs/bronto.grs"))));
      scene.addPosition(new Position(new ObjSimpleModel(
                new File(assets + "cow.obj"))));
      // parametric curves and surfaces
      scene.addPosition(new Position(new ParametricCurve()));
      scene.addPosition(new Position(new ParametricSurface()));
      scene.addPosition(new Position(new ParametricSurface(
                (s,t) -> s*Math.cos(t*Math.PI),
                (s,t) -> t,
                (s,t) -> s*Math.sin(t*Math.PI),
                -1, 1, -1, 1, 49, 49)));
      scene.addPosition(new Position(new ParametricSurface(
                (u,v) -> 0.3*(1-u)*(3+Math.cos(v))*Math.sin(4*Math.PI*u),
                (u,v) -> 0.3*(3*u+(1-u)*Math.sin(v)),
                (u,v) -> 0.3*(1-u)*(3+Math.cos(v))*Math.cos(4*Math.PI*u),
                0, 1, 0, 2*Math.PI, 49, 49)));
      scene.addPosition(new Position(new SurfaceOfRevolution()));
      scene.addPosition(new Position(new SurfaceOfRevolution(
                t -> 0.5*(1+t*t), -1, 1, 30, 30)));
      scene.addPosition(new Position(new SurfaceOfRevolution(
                t -> t, t->4*t*(1-t), 0, 1, 30, 30)));
      // coordinate axes
      scene.addPosition(new Position(new Axes3D(1, 1, 1)));
      scene.addPosition(new Position(new Axes2D(-1, 1, -1, 1, 8, 8)));
      scene.addPosition(new Position(new PanelXY(-4, 4, -5, 5, -5)));
      scene.addPosition(new Position(new PanelXZ(-1, 1, -6, 1, -0.5)));
      scene.addPosition(new Position(new PanelYZ(-1, 1, -6, 1, -0.5)));

      // Make the models invisible, except for the current model.
      for (Position p : scene.positionList)
      {
         p.visible = false;
      }
      currentModel = 0;
      scene.getPosition(currentModel).visible = true;

      // Have the models pushed away from where the camera is.
      deltaX =  0;
      deltaY =  0;
      deltaZ = -2;
      scene.getPosition(currentModel).translation(deltaX, deltaY, deltaZ);

      Rasterize.doClipping = true;


      // Define initial dimensions for a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;

      // Create a FrameBufferPanel that holds a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 2");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setLocationRelativeTo(null);
      jf.setVisible(true);

      // Register this object as the event listener for JFrame events.
      jf.addKeyListener(this);
      jf.addComponentListener(this);

      print_help_message();
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e){}
   @Override public void keyReleased(KeyEvent e){}
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c && e.isAltDown())
      {
         System.out.println();
         System.out.println(scene.getPosition(currentModel).getModel());
      }
      else if ('d' == c)
      {
         scene.debug = ! scene.debug;
      }
      else if ('D' == c)
      {
         Rasterize.debug = ! Rasterize.debug;
      }
      else if ('c' == c)
      {
         Rasterize.doClipping = ! Rasterize.doClipping;
         System.out.print("Clipping is turned ");
         System.out.println(Rasterize.doClipping ? "On" : "Off");
      }
      else if ('/' == c)
      {
         scene.getPosition(currentModel).visible = false;
         currentModel = (currentModel + 1) % scene.positionList.size();
         scene.getPosition(currentModel).visible = true;
         savedModel = null;
         pointSize = 0;
      }
      else if ('?' == c)
      {
         scene.getPosition(currentModel).visible = false;
         currentModel = currentModel - 1;
         if (currentModel < 0) currentModel = scene.positionList.size() - 1;
         scene.getPosition(currentModel).visible = true;
         savedModel = null;
         pointSize = 0;
      }
      else if ('p' == c)
      {
         scene.getCamera().perspective = ! scene.getCamera().perspective;
         final String p = scene.getCamera().perspective ? "perspective"
                                                        : "orthographic";
         System.out.println("Using " + p + " projection");
      }
      else if ('P' == c)
      {
         if (savedModel != null)
         {
            scene.getPosition(currentModel).setModel(savedModel);
            savedModel = null;
            ++pointSize;
         }
         else
         {
            final Model model = scene.getPosition(currentModel).getModel();
            savedModel = model;
            scene.getPosition(currentModel)
                    .setModel(PointCloud.make(model, pointSize));
         }
      }
      else if ('>' == c || '<' == c
             ||'.' == c || ',' == c)
      {
         final Position p = scene.getPosition(currentModel);
         if (p.getModel() instanceof MeshMaker)
         {
            final MeshMaker model = (MeshMaker)(p.getModel());
            final int n;
            final int k;
            if ('>' == c)
            {
               n = model.getHorzCount() + 1;
               k = model.getVertCount();
            }
            else if ('<' == c)
            {
               n = model.getHorzCount() - 1;
               k = model.getVertCount();
            }
            else if ('.' == c)
            {
               n = model.getHorzCount();
               k = model.getVertCount() + 1;
            }
            else  // ',' == c
            {
               n = model.getHorzCount();
               k = model.getVertCount() - 1;
            }
            try
            {
               p.setModel( model.remake(n, k) );
               final MeshMaker m = (MeshMaker)(p.getModel());
               final int newN = m.getHorzCount();
               final int newK = m.getVertCount();
               System.out.printf("Mesh: n = %3d and k = %3d.\n", newN, newK);
            }
            catch (IllegalArgumentException ex){}
         }
      }
      else if ('m' == c) // Display translation information.
      {
         displayTranslation = ! displayTranslation;
      }
      else if ('=' == c) // Reset the translation vector.
      {
         deltaX =  0;
         deltaY =  0;
         deltaZ = -2;
      }
      else if ('x' == c) // Translate ALL the models.
      {
         deltaX += -0.1;  // left
      }
      else if ('X' == c)
      {
         deltaX += +0.1;  // right
      }
      else if ('y' == c)
      {
         deltaY += -0.1;  // down
      }
      else if ('Y' == c)
      {
         deltaY += +0.1;  // up
      }
      else if ('z' == c)
      {
         deltaZ += -0.1;  // back
      }
      else if ('Z' == c)
      {
         deltaZ += +0.1;  // forward
      }
      else if ('+' == c)
      {
         takeScreenshot = true;
      }

      scene.getPosition(currentModel).translation(deltaX, deltaY, deltaZ);

      if (displayTranslation && ('m'==c||'='==c
                               ||'x'==c||'y'==c||'z'==c
                               ||'X'==c||'Y'==c||'Z'==c))
      {
         System.out.printf("deltaX = % .8f, deltaY = % .8f, " +
                           "deltaZ = % .8f\n",
                            deltaX, deltaY, deltaZ);
      }

      // Render again.
      final FrameBuffer fb = fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb);
      if (takeScreenshot)
      {
         fb.dumpFB2File(String.format("Screenshot%03d.png", screenshotNumber),
                        "png");
         ++screenshotNumber;
         takeScreenshot = false;
      }
      fbp.update();
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        fbp.getTopLevelAncestor().getWidth(),
                        fbp.getTopLevelAncestor().getHeight(),
                        fbp.getWidth(), fbp.getHeight());

      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final FrameBuffer fb = new FrameBuffer(w, h);
      fbp.setFrameBuffer(fb);
      Pipeline.render(scene, fb);
      fbp.update();
   }


   private void print_help_message()
   {
      System.out.println("Use the 'd/D' keys to toggle debugging information on and off for the current model.");
      System.out.println("Use the '/' and '?' keys to cycle forwards and backwards through the models.");
      System.out.println("Use the '>/<' and shift keys to increase and decrease the mesh divisions in each direction.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the models along the x, y, z axes.");
      System.out.println("Use the 'm' key to toggle the display of translation information.");
      System.out.println("Use the '=' key to reset the model translation.");
      System.out.println("Use the 'c' key to toggle line clipping on and off.");
      System.out.println("Use the 'P' key to convert the current model to a point cloud.");
      System.out.println("Use the '+' key to save a \"screenshot\" of the framebuffer.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new InteractiveModelsAll_R2()
      );
   }
}
