/*
 * Renderer 2. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.pipeline;

import renderer.scene.*;

import java.util.List;
import java.util.ArrayList;

/**
   Transform each {@link Vertex} of a {@link Model} from the model's
   (private) local coordinate system to the {@link Camera}'s (shared)
   coordinate system.
<p>
   For each {@code Vertex} object in a {@code Model} object, use a
   {@link Position}'s translation {@link Vector} to translate the
   object's {@code Vertex} coordinates from the model's coordinate
   system to the camera's coordinate system.
*/
public class Model2Camera
{
   /**
      Use a {@link Position}'s translation {@link Vector} to transform
      each {@link Vertex} from a {@link Model}'s coordinate system to
      the {@link Camera}'s coordinate system.

      @param position  {@link Position} with a {@link Model} and a translation {@link Vector}
      @return a new {@link Model} with {@link Vertex} objects in the camera's coordinate system
   */
   public static Model model2camera(final Position position)
   {
      final Model model = position.getModel();
      final Vector translate = position.getTranslation();

      // A new vertex list to hold the translated vertices.
      final List<Vertex> newVertexList =
                            new ArrayList<>(model.vertexList.size());

      // Replace each Vertex object with one that
      // contains camera coordinates.
      for (final Vertex v : model.vertexList)
      {
         newVertexList.add( translate.plus(v) );
      }

      return new Model(newVertexList,
                       model.primitiveList,
                       position.getName() + "::" + model.getName(),
                       model.visible);
   }



   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private Model2Camera() {
      throw new AssertionError();
   }
}
