/*
 * Renderer 2. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.scene;

import renderer.scene.primitives.*;

import java.util.List;
import java.util.ArrayList;

/**
   A {@code Model} object represents a distinct geometric object in a
   {@link Scene}. A {@code Model} data structure is mainly a {@link List}
   of {@link Vertex} objects and another {@link List} of {@link Primitive}
   objects.
<p>
   Each {@link Vertex} object contains the xyz-coordinates, in the
   {@code Model}'s own coordinate system, for one point from the
   {@code Model}.
<p>
   The {@link Vertex} objects represents points from the geometric object
   that we are modeling. In the real world, a geometric object has an infinite
   number of points. In 3D graphics, we "approximate" a geometric object by
   listing just enough points to adequately describe the object. For example,
   in the real world, a rectangle contains an infinite number of points, but
   it can be adequately modeled by just its four corner points. (Think about
   a circle. How many points does it take to adequately model a circle? Look
   at the {@link renderer.models_L.Circle} model.)
<p>
   Each {@link Primitive} object is either a {@link LineSegment} or a
   {@link Point}.
<p>
   Each {@link LineSegment} object contains two integers that are the indices
   of two {@link Vertex} objects from the {@code Model}'s vertex list. The
   two vertices are the endpoints of the line segment.
<p>
   Each {@link Point} object contains one integer index of a {@link Vertex}
   from the {@code Model}'s vertex list.
<p>
   We use {@link LineSegment} objects to represent the space between the
   model's vertices. For example, while a rectangle can be approximated by
   its four corner points, those same four points could also represent two
   parallel line segments, or they could represent two lines that cross each
   other. By using four line segments that connect around the four points,
   we get a good, unambiguous representation of a rectangle.
<p>
   If we modeled a circle using just points, we would probably need to use
   hundreds of points. But if we connect every two adjacent points with a
   short line segment, we can get a good model of a circle with just a few
   dozen points.
<p>
   Our {@code Model}'s represent geometric objects as a "wire-frame" of line
   segments, that is, a geometric object is drawn as a collection of "edges".
   This is a fairly simplistic way of doing 3D graphics and we will improve
   this in later renderers.
<p>
   See
<br> <a href="https://en.wikipedia.org/wiki/Wire-frame_model" target="_top">
              https://en.wikipedia.org/wiki/Wire-frame_model</a>
<br>or
<br> <a href="https://www.google.com/search?q=computer+graphics+wireframe&tbm=isch" target="_top">
              https://www.google.com/search?q=computer+graphics+wireframe&tbm=isch</a>
*/
public class Model
{
   public final List<Vertex> vertexList;
   public final List<Primitive> primitiveList;

   protected String name;
   public boolean visible;

   /**
      Construct an empty {@code Model} object.
   */
   public Model()
   {
      this(new ArrayList<>(),  // vertexList
           new ArrayList<>(),  // primitiveList
           "",                 // name
           true);              // visible
   }


   /**
      Construct an empty {@code Model} object with the given name.

      @param name  a {link String} that is a name for this {@code Model}
      @throws NullPointerException if {@code name} is {@code null}
   */
   public Model(final String name)
   {
      this(new ArrayList<>(),  // vertexList
           new ArrayList<>(),  // primitiveList
           name,               // name
           true);              // visible
   }


   /**
      Construct a {@code Model} object with all the given data.

      @param vertexList     a {@link Vertex} {link List} for this {@code Model}
      @param primitiveList  a {@link Primitive} {link List} for this {@code Model}
      @param name           a {link String} that is a name for this {@code Model}
      @param visible        a {@code boolean} that determines this {@code Model}'s visibility
      @throws NullPointerException if {@code vertexList} is {@code null}
      @throws NullPointerException if {@code primitiveList} is {@code null}
      @throws NullPointerException if {@code name} is {@code null}
      @throws NullPointerException if {@code name} is {@code null}
   */
   public Model(final List<Vertex> vertexList,
                final List<Primitive> primitiveList,
                final String name,
                final boolean visible)
   {
      if (null == vertexList)
         throw new NullPointerException("vertexList must not be null");
      if (null == primitiveList)
         throw new NullPointerException("primitiveList must not be null");
      if (null == name)
         throw new NullPointerException("name must not be null");

      this.vertexList = vertexList;
      this.primitiveList = primitiveList;
      this.name = name;
      this.visible = visible;
   }


   /**
      Get a reference to this {@code Model}'s name.

      @return a reference to this {@code Model}'s name
   */
   public String getName()
   {
      return this.name;
   }


   /**
      Set this {@code Model}'s name.

      @param name  {link String} that is the new name for this {@code Model}
      @throws NullPointerException if {@code name} is {@code null}
   */
   public void setName(String name)
   {
      if (null == name)
         throw new NullPointerException("name must not be null");

      this.name = name;
   }


   /**
      Add a {@link Vertex} (or vertices) to this {@code Model}'s
      {@link List} of vertices.

      @param vArray  array of {@link Vertex} objects to add to this {@code Model}
   */
   public final void addVertex(final Vertex... vArray)
   {
      for (final Vertex v : vArray)
      {
         vertexList.add( v );
      }
   }


   /**
      Get a {@link Primitive} from this {@code Model}'s
      {@link List} of primitives.

      @param index  integer index of a {@link Primitive} from this {@code Model}
      @return the {@link Primitive} object at the given index
   */
   public final Primitive getPrimitive(final int index)
   {
      return primitiveList.get(index);
   }


   /**
      Add a {@link Primitive} (or Primitives) to this {@code Model}'s
      {@link List} of primitives.
      <p>
      NOTE: This method does not add any vertices to the {@code Model}'s
      {@link Vertex} list. This method assumes that the appropriate vertices
      have been added to the {@code Model}'s {@link Vertex} list.

      @param pArray  array of {@link Primitive} objects to add to this {@code Model}
   */
   public final void addPrimitive(final Primitive... pArray)
   {
      for (final Primitive p : pArray)
      {
         primitiveList.add(p);
      }
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Model} object
   */
   @Override
   public String toString()
   {
      String result = "";
      result += "Model: " + name + "\n";
      result += "This Model's visibility is: " + visible + "\n";
      result += "Model has " + vertexList.size() + " vertices.\n";
      result += "Model has " + primitiveList.size() + " primitives.\n";
      int i = 0;
      for (final Vertex v : this.vertexList)
      {
         result += i + ": " + v.toString() + "\n";
         ++i;
      }
      i = 0;
      for (final Primitive p : this.primitiveList)
      {
         result += i + ": " + p.toString() + "\n";
         ++i;
      }
      return result;
   }
}
