/*
 * Renderer 2. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.scene;

/**
   A {@code Position} data structure represents a geometric object
   in a distinct location in three-dimensional camera space as part
   of a {@link Scene}.
<p>
   A {@code Position} object holds references to a {@link Model} object
   and a {@link Vector} object. The {@link Model} represents the geometric
   object in the {@link Scene}. The {@link Vector} determines the model's
   location in the {@link Camera} coordinate system.
<p>
   When the renderer renders this {@code Position}'s {@link Model} into
   a {@link renderer.framebuffer.FrameBuffer}, the first stage of the
   rendering pipeline, {@link renderer.pipeline.Model2Camera}, adds this
   {@code Position}'s translation {@link Vector} to every {@link Vertex}
   in the {@link Model}'s vertex list, which converts the coordinates in
   each {@link Vertex} from the model's own local coordinate system to
   the {@link Camera}'s coordinate system (which is "shared" by all
   the other models in the scene). This vector addition has the effect
   of "placing" the model in camera space at an appropriate location.
*/
public final class Position
{
   private Model model;
   private Vector translate;
   public final String name;
   public boolean visible;
   public boolean debug;

   /**
      Construct a {@code Position} with the zero translation {@link Vector}
      and the given {@link Model} object.

      @param model  {@link Model} object to place at this {@code Position}
      @throws NullPointerException if {@code model} is {@code null}
   */
   public Position(final Model model)
   {
      this(model,
           model.getName(),     // default Position name
           new Vector(0, 0, 0), // default translation vector
           true,                // visible
           false);              // debug
   }


   /**
      Construct a {@code Position} with the zero translation {@link Vector},
      the given name, and the given {@link Model} object.

      @param model  {@link Model} object to place at this {@code Position}
      @param name   {@link String} name for this {@code Position}
      @throws NullPointerException if {@code model} is {@code null}
      @throws NullPointerException if {@code name} is {@code null}
   */
   public Position(final Model model, final String name)
   {
      this(model,
           name,
           new Vector(0, 0, 0), // default translation vector
           true,                // visible
           false);              // debug
   }


   /**
      Construct a {@code Position} object with all the given data.

      @param model      {@link Model} object to place at this {@code Position}
      @param name       {@link String} name for this {@code Position}
      @param translate  translation {@link Vector} for this {@code Position}
      @param visible    boolean that determines this {@code Position}'s visibility
      @param debug      boolean that determines if this {@code Position} is logged
      @throws NullPointerException if {@code model} is {@code null}
      @throws NullPointerException if {@code translate} is {@code null}
      @throws NullPointerException if {@code name} is {@code null}
   */
   public Position(final Model model,
                   final String name,
                   final Vector translate,
                   final boolean visible,
                   final boolean debug)
   {
      if (null == model)
         throw new NullPointerException("model must not be null");
      if (null == name)
         throw new NullPointerException("name must not be null");
      if (null == translate)
         throw new NullPointerException("translate must not be null");

      this.model = model;
      this.translate = translate;
      this.name = name;
      this.visible = visible;
      this.debug = debug;
   }


   /**
      Get a reference to this {@code Position}'s name.

      @return a reference to this {@code Position}'s name
   */
   public String getName()
   {
      return this.name;
   }


   /**
      Get a reference to this {@code Position}'s {@link Model} object.

      @return a reference to this {@code Position}'s {@link Model} object
   */
   public Model getModel()
   {
      return this.model;
   }


   /**
      Set this {@code Position}'s {@link Model} object.

      @param model  {@link Model} object to place at this {@code Position}
      @return a reference to the new {@link Model} object to facilitate chaining method calls
      @throws NullPointerException if {@code model} is {@code null}
   */
   public Model setModel(final Model model)
   {
      if (null == model)
         throw new NullPointerException("model must not be null");

      this.model = model;
      return this.model;
   }


   /**
      Get a reference to this {@code Position}'s {@link Vector} object.

      @return a reference to this {@code Position}'s {@link Vector} object
   */
   public Vector getTranslation()
   {
      return this.translate;
   }


   /**
      Set this {@code Position}'s translation {@link Vector} object.

      @param dx  translation amount in the x-direction
      @param dy  translation amount in the y-direction
      @param dz  translation amount in the z-direction
   */
   public void translation(final double dx,
                           final double dy,
                           final double dz)
   {
      this.translate = new Vector(dx, dy, dz);
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Position} object
   */
   @Override
   public String toString()
   {
      String result = "";
      result += "Position: " + name + "\n";
      result += "This Position's visibility is: " + visible + "\n";
      result += "This Position's translation is\n";
      result += translate + "\n";
      result += "This Position's Model is\n";
      result += (null == model) ? "null\n" : model;
      return result;
   }
}
